/* ArduinoFloppyReader (and writer)
*
* Copyright (C) 2017-2022 Robert Smith (@RobSmithDev)
* https://amiga.robsmithdev.co.uk
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Library General Public
* License as published by the Free Software Foundation; either
* version 3 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Library General Public License for more details.
*
* You should have received a copy of the GNU Library General Public
* License along with this library; if not, see http://www.gnu.org/licenses/
*/

//////////////////////////////////////////////////////////////////////////////////////////
// Class that reads reads the MFM data, decodes it, and then writes an ADF file to disk //
//////////////////////////////////////////////////////////////////////////////////////////
//
// Purpose:
// This class reads the raw data from the device and attemps to check for errors.
// Once all errors have been corrected the class will save the tracks to disk as an ADF file
//
// The class also handles writing an ADF file back to disk and optionally verifying the write.
// 
// The MFM decoding algorithm and information regarding finding the start of a sector
// were taken from the excellent documentation by Laurent Cl�vy at http://lclevy.free.fr/adflib/adf_info.html
// Also credits to Keith Monahan https://www.techtravels.org/tag/mfm/ regarding a bug in the MFM sector start data
//

#include <vector>
#include <algorithm>
#include <assert.h>
#include <sstream>
#include <fstream>
#include <iostream>
#include <string>
#include <codecvt>
#include <thread>

// This gets around an issue with the windows header files defining max
const long long StreamMax = std::numeric_limits<std::streamsize>::max();

#include "ADFWriter.h"
#include "RotationExtractor.h"
#include <Windows.h>
#include "pll.h"

#include <math.h>


using namespace ArduinoFloppyReader;


#define MFM_MASK    0x55555555L		
#define AMIGA_WORD_SYNC  0x4489							 // Disk SYNC code for the Amiga start of sector
#define SECTOR_BYTES	512								 // Number of bytes in a decoded sector
#define NUM_SECTORS_PER_TRACK_DD 11						 // Number of sectors per track
#define NUM_SECTORS_PER_TRACK_HD 22						  // Same but for HD disks
#define RAW_SECTOR_SIZE (8+56+SECTOR_BYTES+SECTOR_BYTES)      // Size of a sector, *Including* the sector sync word longs
#define ADF_TRACK_SIZE_DD (SECTOR_BYTES*NUM_SECTORS_PER_TRACK_DD)   // Bytes required for a single track dd
#define ADF_TRACK_SIZE_HD (SECTOR_BYTES*NUM_SECTORS_PER_TRACK_HD)   // Bytes required for a single track hd

// Buffer to hold raw data for just a single sector
typedef unsigned char RawEncodedSector[RAW_SECTOR_SIZE];
typedef unsigned char RawDecodedSector[SECTOR_BYTES];
typedef RawDecodedSector RawDecodedTrackDD[NUM_SECTORS_PER_TRACK_DD];
typedef RawDecodedSector RawDecodedTrackHD[NUM_SECTORS_PER_TRACK_HD];
typedef unsigned char RawMFMData[SECTOR_BYTES + SECTOR_BYTES];

// When workbench formats a disk, it write 13630 bytes of mfm data to the disk.  So we're going to write this amount, and then we dont need an erase first
typedef struct alignas(1) {
	unsigned char filler1[1654];  // Padding at the start of the track.  This will be set to 0xaa
	// Raw sector data
	RawEncodedSector sectors[NUM_SECTORS_PER_TRACK_DD];   // 11968 bytes
	// Blank "Filler" gap content. (this may get overwritten by the sectors a little)
	unsigned char filler2[8];
} FullDiskTrackDD;

typedef struct alignas(1) {
	unsigned char filler1[1654];  // Padding at the start of the track.  This will be set to 0xaa
	// Raw sector data
	RawEncodedSector sectors[NUM_SECTORS_PER_TRACK_HD];   // 11968*2 bytes
	// Blank "Filler" gap content. (this may get overwritten by the sectors a little)
	unsigned char filler2[8];
} FullDiskTrackHD;

// Structure to hold data while we decode it
typedef struct alignas(8)  {
	unsigned char trackFormat;        // This will be 0xFF for Amiga
	unsigned char trackNumber;        // Current track number (this is actually (tracknumber*2) + side
	unsigned char sectorNumber;       // The sector we just read (0 to 11)
	unsigned char sectorsRemaining;   // How many more sectors remain until the gap (0 to 10)

	uint32_t sectorLabel[4];     // OS Recovery Data, we ignore this

	uint32_t headerChecksum;	  // Read from the header, header checksum
	uint32_t dataChecksum;		  // Read from the header, data checksum

	uint32_t headerChecksumCalculated;   // The header checksum we calculate
	uint32_t dataChecksumCalculated;     // The data checksum we calculate

	RawDecodedSector data;          // decoded sector data

	RawMFMData rawSector;   // raw track data, for analysis of invalid sectors
} DecodedSector;

// To hold a list of valid and checksum failed sectors
struct DecodedTrack {
	// A list of valid sectors where the checksums are OK
	std::vector<DecodedSector> validSectors;
	// A list of sectors found with invalid checksums.  These are used if ignore errors is triggered
	// We keep copies of each one so we can perform a statistical analysis to see if we can get a working one based on which bits are mostly set the same
	std::vector<DecodedSector> invalidSectors[NUM_SECTORS_PER_TRACK_HD];
};


// MFM decoding algorithm
// *input;	MFM coded data buffer (size == 2*data_size) 
// *output;	decoded data buffer (size == data_size) 
// Returns the checksum calculated over the data
uint32_t decodeMFMdata(const uint32_t* input, uint32_t* output, const unsigned int data_size) {
	uint32_t odd_bits, even_bits;
	uint32_t chksum = 0L;
	unsigned int count;

	// the decoding is made here long by long : with data_size/4 iterations 
	for (count = 0; count < data_size / 4; count++) {
		odd_bits = *input;					// longs with odd bits 
		even_bits = *(uint32_t*)(((unsigned char*)input) + data_size);   // longs with even bits - located 'data_size' bytes after the odd bits

		chksum ^= odd_bits;              // XOR Checksum
		chksum ^= even_bits;

		*output = ((even_bits & MFM_MASK) | ((odd_bits & MFM_MASK) << 1));
		input++;      /* next 'odd' long and 'even bits' long  */
		output++;     /* next location of the future decoded long */
	}
	return chksum & MFM_MASK;
}

// Copys the data from inTrack into outTrack but fixes the bit/byte alignment so its aligned on the start of a byte 
void alignSectorToByte(const unsigned char* inTrack, const int dataLength, int byteStart, int bitStart, RawEncodedSector& outSector) {
	unsigned char byteOut = 0;
	unsigned int byteOutPosition = 0;

	// Bit counter output
	unsigned int counter = 0;

	// The position supplied is the last bit of the track sync.  
	bitStart--;   // goto the next bit
	if (bitStart < 0) {
		// Could do a MEMCPY here, but for now just let the code below run
		bitStart = 7;
		byteStart++;
	}
	byteStart -= 8;   // wind back 8 bytes

	// This is mis-aligned.  So we need to shift the data into byte boundarys
	for (;;) {
		for (int bitCounter = bitStart; bitCounter >= 0; bitCounter--) {
			byteOut <<= 1;
			if (inTrack[byteStart % dataLength] & (1 << bitCounter)) byteOut |= 1;

			if (++counter >= 8) {
				outSector[byteOutPosition] = byteOut;
				byteOutPosition++;
				if (byteOutPosition >= RAW_SECTOR_SIZE) return;
				counter = 0;
			}
		}

		// Move along and reset
		byteStart++;
		bitStart = 7;
	}
}

// Attempt to repair the MFM data.  Returns TRUE if errors are detected
bool repairMFMData(unsigned char* data, const unsigned int dataLength) {
	bool errors = false;
	// Only certain bit-patterns are allowed.  So if we come across an invalid one we will try to repair it.  
	// You cannot have two '1's together, and a max of three '0' in a row
	// Allowed combinations:  (note the SYNC WORDS and TRACK START are designed to break these rules, but we shouldn't encounter them)
	// 
	//	00010
	//	00100
	//	00101
	//	01000
	//	01001
	//	01010
	//	10001
	//	10010
	//	10100
	//	10101
	//
	unsigned char testByte = 0;
	int counter = 0;
	for (unsigned int position = 0; position < dataLength; position++) {
		// Fixed: This was the wrong way around
		for (int bitIndex = 7; bitIndex >= 0; bitIndex--) {
			testByte <<= 1;   // shift off one bit to make room for the new bit
			if (*data & (1 << bitIndex)) {
				// Make sure two '1's dont come in together as this is not allowed! This filters out a lot of BAD combinations
				if ((testByte & 0x2) != 0x2) {
					testByte |= 1;
				} 
				else {
					// We detected two '1's in a row, which isnt allowed.  From reading this most likely means this was a weak bit, so we change it to zero.
					errors = true;
				}
			} 

			// We're only interested in the last so many bits, and only when we have received that many
			if (++counter > 4) {
				switch (testByte & 0x1F) {
					// These are the only possible invalid combinations left
				case 0x00:
				case 0x01:
				case 0x10:
					// No idea how to repair these	
					errors = true;
					break;
				}
			}
		}
		data++;
	}

	return errors;
	 
}

// Looks at the history for this sector number and creates a new sector where the bits are set to whatever occurs more.  We then do a checksum and if it succeeds we use it
bool attemptFixSector(const DecodedTrack& decodedTrack, DecodedSector& outputSector) {
	int sectorNumber = outputSector.sectorNumber;

	if (decodedTrack.invalidSectors[sectorNumber].size() < 2) return false;

	typedef struct {
		int zeros = 0;
		int ones = 0;
	} SectorCounter[8];

	SectorCounter sectorSum[SECTOR_BYTES + SECTOR_BYTES];

	memset(sectorSum, 0, sizeof(sectorSum));

	// Calculate the number of '1's and '0's in each block
	for (const DecodedSector& sec : decodedTrack.invalidSectors[sectorNumber]) 
		for (int byteNumber = 0; byteNumber < SECTOR_BYTES + SECTOR_BYTES; byteNumber++) 
			for (int bit = 0; bit <= 7; bit++) 
				if (sec.rawSector[byteNumber] & (1 << bit))
					sectorSum[byteNumber][bit].ones++; else sectorSum[byteNumber][bit].zeros++;

	// Now create a sector based on this data
	memset(outputSector.rawSector, 0, sizeof(outputSector.rawSector));
	for (int byteNumber = 0; byteNumber < SECTOR_BYTES + SECTOR_BYTES; byteNumber++)
		for (int bit = 0; bit <= 7; bit++)
			if (sectorSum[byteNumber][bit].ones >= sectorSum[byteNumber][bit].zeros)
				outputSector.rawSector[byteNumber] |= (1 << bit);

	return true;
}

// Extract and convert the sector.  This may be a duplicate so we may reject it.  Returns TRUE if it was valid, or false if not
bool decodeSector(const RawEncodedSector& rawSector, const unsigned int trackNumber, bool isHD, const DiskSurface surface, DecodedTrack& decodedTrack, bool ignoreHeaderChecksum, int& lastSectorNumber) {
	DecodedSector sector;

	lastSectorNumber = -1;
	memcpy(sector.rawSector, rawSector, sizeof(RawMFMData));

	// Easier to operate on
	unsigned char* sectorData = (unsigned char*)rawSector;
 
	// Read the first 4 bytes (8).  This  is the track header data	
	sector.headerChecksumCalculated = decodeMFMdata((uint32_t*)(sectorData + 8), (uint32_t*)&sector, 4);
	// Decode the label data and update the checksum
	sector.headerChecksumCalculated ^= decodeMFMdata((uint32_t*)(sectorData + 16), (uint32_t*)&sector.sectorLabel[0], 16);
	// Get the checksum for the header
	decodeMFMdata((uint32_t*)(sectorData + 48), (uint32_t*)&sector.headerChecksum, 4);  // (computed on mfm longs, longs between offsets 8 and 44 == 2 * (1 + 4) longs)
	// If the header checksum fails we just cant trust anything we received, so we just drop it
	if ((sector.headerChecksum != sector.headerChecksumCalculated) && (!ignoreHeaderChecksum)) {
		return false;
	}

	// Check if the header contains valid fields
	if (sector.trackFormat != 0xFF) 
		return false;  // not valid
	if (sector.sectorNumber > (isHD ? 21 : 10))
		return false;
	if (sector.trackNumber > 166) 
		return false;   // 83 tracks * 2 for both sides
	if (sector.sectorsRemaining > (isHD ? 22 : 11))
		return false;  // this isnt possible either
	if (sector.sectorsRemaining < 1)
		return false;  // or this

	// And is it from the track we expected?
	const unsigned char targetTrackNumber = (trackNumber << 1) | ((surface == DiskSurface::dsUpper) ? 1 : 0);

	if (sector.trackNumber != targetTrackNumber) return false; // this'd be weird

	// Get the checksum for the data
	decodeMFMdata((uint32_t*)(sectorData + 56), (uint32_t*)&sector.dataChecksum, 4);
	

	// Lets see if we already have this one
	const int searchSector = sector.sectorNumber;
	auto index = std::find_if(decodedTrack.validSectors.begin(), decodedTrack.validSectors.end(), [searchSector](const DecodedSector& sector) -> bool {
		return (sector.sectorNumber == searchSector);
	});

	// We already have it as a GOOD VALID sector, so skip, we don't need it.
	if (index != decodedTrack.validSectors.end()) return true;

	// Decode the data and receive it's checksum
	sector.dataChecksumCalculated = decodeMFMdata((uint32_t*)(sectorData + 64), (uint32_t*)&sector.data[0], SECTOR_BYTES); // (from 64 to 1088 == 2*512 bytes)

	lastSectorNumber = sector.sectorNumber;

	// Is the data valid?
	if (sector.dataChecksum != sector.dataChecksumCalculated) {
		// Keep a copy
		decodedTrack.invalidSectors[sector.sectorNumber].push_back(sector);
		return false;
	}
	else {
		// Its a good sector, and we dont have it yet
		decodedTrack.validSectors.push_back(sector);

		// Lets delete it from invalid sectors list
		decodedTrack.invalidSectors[sector.sectorNumber].clear();


		return true;
	}
}

// Find sectors within raw data read from the drive by searching bit-by-bit for the SYNC bytes
void findSectors(const unsigned char* track, bool isHD, unsigned int trackNumber, DiskSurface side, unsigned short trackSync, DecodedTrack& decodedTrack, bool ignoreHeaderChecksum) {
	// Work out what we need to search for which is syncsync
	const uint32_t search = (trackSync | (((uint32_t)trackSync) << 16));

	// Prepare our test buffer
	uint32_t decoded = 0;

	// Keep runnign until we run out of data
	unsigned int byteIndex = 0;

	int nextTrackBitCount = 0;

	const unsigned int dataLength = isHD ? RAW_TRACKDATA_LENGTH_HD : RAW_TRACKDATA_LENGTH_DD;
	const int maxSectors = isHD ? NUM_SECTORS_PER_TRACK_HD : NUM_SECTORS_PER_TRACK_DD;

	// run the entire track length
	while (byteIndex < dataLength) {

		// Check each bit, the "decoded" variable slowly slides left providing a 32-bit wide "window" into the bitstream
		for (int bitIndex = 7; bitIndex >= 0; bitIndex--) {
			decoded <<= 1;   // shift off one bit to make room for the new bit

			if (track[byteIndex] & (1 << bitIndex)) decoded |= 1;

			// Have we matched the sync words correctly
			++nextTrackBitCount;
			int lastSectorNumber = -1;
			if (decoded == search) {
				RawEncodedSector alignedSector;
				
				// We extract ALL of the track data from this BIT to byte align it properly, then pass it onto the code to read the sector (from the start of the sync code)
				alignSectorToByte(track, dataLength, byteIndex, bitIndex, alignedSector);

				// Now see if there's a valid sector there.  We now only skip the sector if its valid, incase rogue data gets in there
				if (decodeSector(alignedSector, trackNumber, isHD, side, decodedTrack, ignoreHeaderChecksum, lastSectorNumber)) {
					// We know the size of this buffer, so we can skip by exactly this amount
					byteIndex += RAW_SECTOR_SIZE - 8; // skip this many bytes as we know this is part of the track! minus 8 for the SYNC
					if (byteIndex >= dataLength) break;
					// We know that 8 bytes from here should be another track. - we allow 1 bit either way for slippage, but this allows an extra check incase the SYNC pattern is damaged
					nextTrackBitCount = 0;
				}
				else {

					// Decode failed.  Lets try a "homemade" one
					DecodedSector newTrack;
					if ((lastSectorNumber >= 0) && (lastSectorNumber < maxSectors)) {
						newTrack.sectorNumber = lastSectorNumber;
						if (attemptFixSector(decodedTrack, newTrack)) {
							memcpy(newTrack.rawSector, alignedSector, sizeof(newTrack.rawSector));
							// See if our makeshift data will decode or not
							if (decodeSector(alignedSector, trackNumber, isHD, side, decodedTrack, ignoreHeaderChecksum, lastSectorNumber)) {
								// We know the size of this buffer, so we can skip by exactly this amount
								byteIndex += RAW_SECTOR_SIZE - 8; // skip this many bytes as we know this is part of the track! minus 8 for the SYNC
								if (byteIndex >= dataLength) break;
							}
						}
					}
					if (decoded == search) nextTrackBitCount = 0;
				}
			}
		}
		byteIndex++;
	}
}

// Merges any invalid sectors into the valid ones as a last resort
void mergeInvalidSectors(DecodedTrack& track, bool isHD) {
	const int maxSectors = isHD ? NUM_SECTORS_PER_TRACK_HD : NUM_SECTORS_PER_TRACK_DD;

	for (unsigned char sector = 0; sector < maxSectors; sector++) {
		if (track.invalidSectors[sector].size()) {
			// Lets try to make the best sector we can
			DecodedSector sec = track.invalidSectors[sector][0];
			// Repair maybe!?
			attemptFixSector(track, sec);

			track.validSectors.push_back(sec);
		}
		track.invalidSectors[sector].clear();
	}
}

ADFWriter::ADFWriter() {
}

ADFWriter::~ADFWriter() {
}

bool ADFWriter::diskClean(std::function<void(int progress)> onProgress) {
	if (!m_device.isOpen()) return false;
	if (m_device.enableMotor(true) != GWResponse::drOK) return false;
	const int trk[20] = { 81,10,71,20,61,30,51,70,39,40,41,21,50,31,60,21,70,11,80,0 };
	onProgress(0);
	int progress = 0;
	for (int retry=0; retry<2; retry++) {
		if (m_device.selectTrack(0, TrackSearchSpeed::tssSlow) != GWResponse::drOK) return false;
		onProgress(progress / 2); progress++;
		for (int i = 0; i < 20; i++) {
			onProgress(progress / 2); progress++;
			if (m_device.selectTrack(trk[i], TrackSearchSpeed::tssSlow) != GWResponse::drOK) return false;
			for (int j = 1; j <= 4; j++) {
				onProgress(progress / 2); progress++;
				Sleep(50);
			}
		}
	}
	if (m_device.enableMotor(false) != GWResponse::drOK) return false;
	return true;
}

// Open the device we want to use.  Returns TRUE if it worked
bool ADFWriter::openDevice(const std::wstring& portName, bool useDriveB) {
	return m_device.openPort(portName, useDriveB) == GWResponse::drOK;
}

// Close the device when we've finished
void ADFWriter::closeDevice() {
	m_device.closePort();
}


bool ADFWriter::fetchDiskLabel(bool inHDMode, std::string& label) {
	label = "";
	if (!m_device.isOpen()) return false;

	const unsigned int readSize = m_isInHDMode ? sizeof(RawTrackDataHD) : sizeof(RawTrackDataDD);
	const unsigned int maxSectorsPerTrack = m_isInHDMode ? NUM_SECTORS_PER_TRACK_HD : NUM_SECTORS_PER_TRACK_DD;

	if (m_device.selectTrack(40) != GWResponse::drOK) {
		return false;
	}
	if (m_device.selectSurface(DiskSurface::dsLower) != GWResponse::drOK) {
		return false;
	}	
	m_device.setDiskCapacity(inHDMode);


	DecodedTrack track;
	for (unsigned int sector = 0; sector < maxSectorsPerTrack; sector++) track.invalidSectors[sector].clear();
	track.validSectors.clear();

	unsigned char* data = (unsigned char*)malloc(RAW_TRACKDATA_LENGTH_HD);
	if (!data) {
		return false;
	}
	memset(data, 0, sizeof(RawTrackDataHD));
	m_device.readCurrentTrack(data, readSize*8);
	findSectors(data, m_isInHDMode, 40, DiskSurface::dsLower, AMIGA_WORD_SYNC, track, true);
	free(data);

	DecodedSector sector;
	sector.trackFormat = 0;
	for (DecodedSector& s : track.validSectors)
		if (s.sectorNumber == 0) {
			sector = s;
			break;
		}
	if (sector.trackFormat == 0) {
		for (DecodedSector& s : track.invalidSectors[0])
			if (s.sectorNumber == 0) {
				sector = s;
				break;
			}
	}
	if (sector.trackFormat != 0) {
		unsigned char* writebuffer = sector.data;
		if (writebuffer[0] == 0 && writebuffer[1] == 0 && writebuffer[2] == 0 && writebuffer[3] == 2 &&
			writebuffer[508] == 0 && writebuffer[509] == 0 && writebuffer[510] == 0 && writebuffer[511] == 1) {
			writebuffer[512 - 20 * 4 + 1 + writebuffer[512 - 20 * 4]] = 0;

			char* n = (char*)(writebuffer + 512 - 20 * 4 + 1);
			label.resize(strlen(n));
			memcpy((void*)label.data(), n, label.size());
		}
	}
	return true;
}

// Reads the disk and write the data to the ADF file supplied.  The callback is for progress, and you can returns FALSE to abort the process
ADFResult ADFWriter::DiskToADF(const std::wstring& outputFile, const bool inHDMode, const unsigned int numTracks, std::function < WriteResponse(const int currentTrack, const DiskSurface currentSide, const int retryCounter, const int sectorsFound, const int badSectorsFound, const int maxSectors, const CallbackOperation operation)> callback) {
	if (!m_device.isOpen()) return ADFResult::adfrDriveError;

	if (callback)
		if (callback(0, DiskSurface::dsLower, 0, 0, 0, 0, CallbackOperation::coStarting) == WriteResponse::wrAbort) return ADFResult::adfrAborted;

	// Higher than this is not supported
	if (numTracks > 84) return ADFResult::adfrDriveError;

	// Attempt ot open the file
#ifdef _WIN32	
	std::ofstream hADFFile = std::ofstream(outputFile, std::ofstream::out | std::ofstream::binary | std::ofstream::trunc);
#else
	std::string outputFileA;
	quickw2a(outputFile, outputFileA);
	std::ofstream hADFFile = std::ofstream(outputFileA, std::ofstream::out | std::ofstream::binary | std::ofstream::trunc);
#endif	
	m_device.setDiskCapacity(inHDMode);

	if (!hADFFile.is_open()) return ADFResult::adfrFileError;

	// To hold a raw track
	RawTrackDataHD data;
	DecodedTrack track;
	bool includesBadSectors = false;

	const unsigned int readSize = inHDMode ? sizeof(RawTrackDataHD) : sizeof(RawTrackDataDD);
	const unsigned int maxSectorsPerTrack = inHDMode ? NUM_SECTORS_PER_TRACK_HD : NUM_SECTORS_PER_TRACK_DD;

	// Do all tracks
	for (unsigned int currentTrack = 0; currentTrack < numTracks; currentTrack++) {

		// Select the track we're working on
		if (m_device.selectTrack(currentTrack) != GWResponse::drOK) {
			hADFFile.close();
			return ADFResult::adfrCompletedWithErrors;
		}

		// Now select the side
		for (unsigned int surfaceIndex = 0; surfaceIndex < 2; surfaceIndex++) {
			// Surface 
			const DiskSurface surface = (surfaceIndex == 1) ? DiskSurface::dsUpper : DiskSurface::dsLower;

			// Change the surface we're looking at
			if (m_device.selectSurface(surface) != GWResponse::drOK) {
				hADFFile.close();
				return ADFResult::adfrCompletedWithErrors;
			}

			// Reset the sectors list
			for (unsigned int sector = 0; sector < maxSectorsPerTrack; sector++)
				track.invalidSectors[sector].clear();

			track.validSectors.clear();

			// Extract phase code
			int failureTotal = 0;
			bool ignoreChecksums = false;

			// Repeat until we have all 11 sectors
			while (track.validSectors.size() < maxSectorsPerTrack) {

				if (callback) {
					int total = 0;
					for (unsigned int sector = 0; sector < maxSectorsPerTrack; sector++)
						if (track.invalidSectors[sector].size()) total++;

					if ((failureTotal % 6) == 5) {
						// simulate what the Amiga kinda sounded like it was doing by re-seeking to the track.  This sometimes fixes it, weird eh, and sounds cool
						if (currentTrack < 40) {
							m_device.selectTrack(currentTrack + 30, TrackSearchSpeed::tssSlow);
						}
						else {
							m_device.selectTrack(currentTrack - 30, TrackSearchSpeed::tssSlow);
						}
						m_device.selectTrack(currentTrack);
					}

					switch (callback(currentTrack, surface, failureTotal, track.validSectors.size(), total, maxSectorsPerTrack, failureTotal > 0 ? CallbackOperation::coRetryReading : CallbackOperation::coReading)) {
					case WriteResponse::wrContinue: break;  // do nothing
					case WriteResponse::wrRetry:    failureTotal = 0; break;
					case WriteResponse::wrAbort:    hADFFile.close();
						return ADFResult::adfrAborted;

					case WriteResponse::wrSkipBadChecksums:
						if (ignoreChecksums) {
							// Already been here, so we'll create blank sectors just to get this going
							for (unsigned char sectornumber = 0; sectornumber <= maxSectorsPerTrack; sectornumber++) {
								auto index = std::find_if(track.validSectors.begin(), track.validSectors.end(), [sectornumber](const DecodedSector& sector) -> bool {
									return (sector.sectorNumber == sectornumber);
									});
								// Not found. Lets add it
								if (index == track.validSectors.end()) {
									DecodedSector sector;
									memset(&sector, 0, sizeof(sector));
									sector.sectorNumber = sectornumber;
									track.validSectors.push_back(sector);
								}
							}
						}
						ignoreChecksums = true;
						failureTotal = 0;
						break;
					}
				}

				if (m_device.readCurrentTrack(data, readSize*8) == GWResponse::drOK) {
					findSectors(data, inHDMode, currentTrack, surface, AMIGA_WORD_SYNC, track, ignoreChecksums);
					failureTotal++;
				}
				else {
					hADFFile.close();
					return ADFResult::adfrDriveError;
				}

				// If the user wants to skip invalid sectors and save them
				if (ignoreChecksums) {
					for (unsigned int sector = 0; sector < maxSectorsPerTrack; sector++)
						if (track.invalidSectors[sector].size()) {
							if (currentTrack<80) 
								includesBadSectors = true;
							break;
						}
					mergeInvalidSectors(track, inHDMode);
				}
			}

			// Sort the sectors in order
			std::sort(track.validSectors.begin(), track.validSectors.end(), [](const DecodedSector& a, const DecodedSector& b) -> bool {
				return a.sectorNumber < b.sectorNumber;
				});

			// Now write all of them to disk
			for (unsigned int sector = 0; sector < maxSectorsPerTrack; sector++) {
				try {
					hADFFile.write((const char*)track.validSectors[sector].data, 512);
				}
				catch (...) {
					hADFFile.close();
					return ADFResult::adfrFileIOError;
				}
			}
		}
	}

	hADFFile.close();

	return includesBadSectors ? ADFResult::adfrCompletedWithErrors : ADFResult::adfrComplete;
}



#pragma pack(1) 

struct SCPFileHeader {
	char			headerSCP[3];
	unsigned char	version;
	unsigned char	diskType;
	unsigned char	numRevolutions;
	unsigned char	startTrack;
	unsigned char   endTrack;
	unsigned char	flags;
	unsigned char	bitcellEncoding;   // 0=16 bits per sample, 
	unsigned char	numHeads;
	unsigned char   timeBase;          // Resolution. Time in ns = (timeBase+1)*25
	uint32_t	checksum;
};

struct SCPTrackHeader {
	char			headerTRK[3];
	unsigned char	trackNumber;
};

struct SCPTrackRevolution {
	uint32_t	indexTime;		// Time in NS/25 for this revolution
	uint32_t	trackLength;	// Number of bit-cells in this revolution
	uint32_t	dataOffset;		// From the start of SCPTrackHeader 
};

// Track data is 16-bit value in NS/25.  If =0 means no flux transition for max time 
#pragma pack()

#define BITFLAG_INDEX		0
#define BITFLAG_96TPI		1
#define BITFLAG_NORMALISED  3
#define BITFLAG_EXTENDED    6
#define BITFLAG_FLUXCREATOR 7

typedef std::vector<uint16_t> SCPTrackData;

struct SCPTrackInMemory {
	SCPTrackHeader header;
	std::vector<SCPTrackRevolution> revolution;
	std::vector<SCPTrackData> revolutionData;
};

// Reads the disk and write the data to the SCP file supplied.  The callback is for progress, and you can returns FALSE to abort the process
ADFResult ADFWriter::DiskToSCP(const std::wstring& outputFile, const unsigned int numTracks, const unsigned char revolutions, float rpm, std::function < WriteResponse(const int currentTrack, const DiskSurface currentSide, const unsigned int trackTimeMS, const CallbackOperation operation)> callback) {
	if (!m_device.isOpen()) return ADFResult::adfrDriveError;

	if (callback)
		if (callback(0, DiskSurface::dsLower, 0, CallbackOperation::coStarting) == WriteResponse::wrAbort) return ADFResult::adfrAborted;

	// Higher than this is not supported
	if (numTracks > 84) return ADFResult::adfrDriveError;
	if (revolutions < 1) return ADFResult::adfrDriveError;
	if (revolutions > 5) return ADFResult::adfrDriveError;

	// Attempt ot open the file
	std::fstream hADFFile = std::fstream(outputFile, std::ofstream::out | std::ofstream::in | std::ofstream::binary | std::ofstream::trunc);
	if (!hADFFile.is_open()) return ADFResult::adfrFileError;

	SCPFileHeader header;
	header.headerSCP[0] = 'S';
	header.headerSCP[1] = 'C';
	header.headerSCP[2] = 'P';
	header.version = 0;// 2 | (2 << 4);    // CHECK
	header.diskType = 0x04; // amiga 0x80;
	header.numRevolutions = revolutions;
	header.startTrack = 0;
	header.numHeads = 0;  // both heads
	header.timeBase = 0;  // 25ns
	header.endTrack = (numTracks * 2) - 1;
	header.flags = (1 << BITFLAG_INDEX) | (1 << BITFLAG_96TPI) | (1 << BITFLAG_FLUXCREATOR);
	header.bitcellEncoding = 0; // 16-bit
	// to be calculated
	header.checksum = 0;

	assert(sizeof(SCPFileHeader) == 16);

	SCPTrackInMemory track;
	track.header.headerTRK[0] = 'T';
	track.header.headerTRK[1] = 'R';
	track.header.headerTRK[2] = 'K';

	try {
		hADFFile.write((const char*)&header, sizeof(header));
	}
	catch (...) {
		hADFFile.close();
		return ADFResult::adfrFileIOError;
	}

	// Pad out the records.  Theres 4 bytes for each track
	uint32_t notPresent = 0;
	for (unsigned int a = 0; a < 168; a++) {
		try {
			hADFFile.write((const char*)&notPresent, sizeof(notPresent));
		}
		catch (...) {
			hADFFile.close();
			return ADFResult::adfrFileIOError;
		}
	}

	// Do all tracks
	for (unsigned int currentTrack = 0; currentTrack < numTracks; currentTrack++) {

		// Select the track we're working on
		if (m_device.selectTrack(currentTrack) != GWResponse::drOK) {
			hADFFile.close();
			return ADFResult::adfrCompletedWithErrors;
		}

		// Now select the side
		for (unsigned int surfaceIndex = 0; surfaceIndex < 2; surfaceIndex++) {

			// Surface 
			const DiskSurface surface = (surfaceIndex == 1) ? DiskSurface::dsUpper : DiskSurface::dsLower;
			track.revolution.clear();
			track.revolutionData.clear();
			track.header.trackNumber = (currentTrack * 2) + surfaceIndex;

			// Change the surface we're looking at
			if (m_device.selectSurface(surface) != GWResponse::drOK) {
				hADFFile.close();
				return ADFResult::adfrCompletedWithErrors;
			}			

			m_device.readCurrentTrackFlux(revolutions, [this , &track, rpm](const std::vector<DWORD>& flux, DWORD extraTimeToAddOn) {
				SCPTrackRevolution currentRev;
				SCPTrackData currentRevData;
				ULONGLONG ttt = extraTimeToAddOn;
				currentRev.trackLength = 0;
				DWORD extra = 0;

				for (DWORD f : flux) {
					// Convert to 25ns times
					//f = (DWORD)(f * 300.0f / rpm);
					if (extra) f+= extra;
					ttt += f;
					extra = f % 25;
					f/= 25;



					// Handle data too big
					while (f > 65535) {
						currentRevData.push_back(0);
						f -= 65536;
					}
					if (f == 0) {
						f = 1;
						extra--;
					}
					currentRevData.push_back((unsigned short)(((f & 0xFF) << 8) | ((f >> 8) & 0xFF)));
					currentRev.trackLength++;
				}
				currentRev.indexTime = ttt / 25;

				track.revolution.push_back(currentRev);
				track.revolutionData.push_back(currentRevData);
			});			
			
			if (track.revolution.size() < revolutions)
				return ADFResult::adfrDriveError;
			if (callback) {
				switch (callback(currentTrack, surface, (unsigned int)((track.revolution[0].indexTime*25ULL)/1000000ULL), CallbackOperation::coReading)) {
				case WriteResponse::wrContinue: break;  // do nothing
				case WriteResponse::wrAbort:    hADFFile.close();
					return ADFResult::adfrAborted;
				default: break;
				}
			}

			// Get the current position
			uint32_t currentPosition = (uint32_t)hADFFile.tellp();

			// Move to the beginning of the file, and write the offset for where this starts
			hADFFile.seekp(sizeof(header) + (track.header.trackNumber * 4), std::fstream::beg);
			try {
				hADFFile.write((const char*)&currentPosition, 4);
			}
			catch (...) {
				hADFFile.close();
				return ADFResult::adfrFileIOError;
			}

			// Restore position and save data
			hADFFile.seekp(currentPosition, std::fstream::beg);

			// Write the header
			try {
				hADFFile.write((const char*)&track.header, sizeof(track.header));
			}
			catch (...) {
				hADFFile.close();
				return ADFResult::adfrFileIOError;
			}

			// Write out the revolution headers
			unsigned int dataPos = sizeof(track.header) + (track.revolution.size() * sizeof(SCPTrackRevolution));
			for (unsigned int a = 0; a < track.revolution.size(); a++) {
				track.revolution[a].dataOffset = dataPos;
				try {
					hADFFile.write((const char*)&track.revolution[a], sizeof(track.revolution[a]));
				}
				catch (...) {
					hADFFile.close();
					return ADFResult::adfrFileIOError;
				}
				dataPos += track.revolutionData[a].size() * 2;
			}

			// Now write out the data
			for (unsigned int a = 0; a < track.revolutionData.size(); a++) {
				try {
					hADFFile.write((const char*)track.revolutionData[a].data(), track.revolutionData[a].size() * 2);
				}
				catch (...) {
					hADFFile.close();
					return ADFResult::adfrFileIOError;
				}
			}
		}
	}

	// Compute the checksum
	hADFFile.seekg(sizeof(SCPFileHeader), std::fstream::beg);
	unsigned char buffer[256];
	header.checksum = 0;

	while (hADFFile.good()) {
		try {
			hADFFile.read((char*)buffer, sizeof(buffer));
			const std::streamsize read = hADFFile.gcount();
			for (size_t pos = 0; pos < (size_t)read; pos++)
				header.checksum += buffer[pos];
		}
		catch (...) {
		}
	}
	hADFFile.clear();
	hADFFile.seekp(0, std::fstream::beg);

	// Write the header again with the checksum in it
	try {
		hADFFile.write((const char*)&header, sizeof(header));
	}
	catch (...) {
		hADFFile.close();
		return ADFResult::adfrFileIOError;
	}

	hADFFile.close();

	return ADFResult::adfrComplete;
}

