/* ArduinoFloppyReader (and writer)
* 
* Copyright (C) 2017-2022 Robert Smith (@RobSmithDev)
* https://amiga.robsmithdev.co.uk
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Library General Public
* License as published by the Free Software Foundation; either
* version 3 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Library General Public License for more details.
*
* You should have received a copy of the GNU Library General Public
* License along with this library; if not, see http://www.gnu.org/licenses/
*/

//////////////////////////////////////////////////////////////////////////////////////////
// Class that reads reads the MFM data, decodes it, and then writes an ADF file to disk //
//////////////////////////////////////////////////////////////////////////////////////////
//
// Purpose:
// This class reads the raw data from the device and attemps to check for errors.
// Once all errors have been corrected the class will save the tracks to disk as an ADF file
//
// The class also handles writing an ADF file back to disk and optionally verifying the write.
// 
// The MFM decoding algorithm and information regarding finding the start of a sector
// were taken from the excellent documentation by Laurent Clvy at http://lclevy.free.fr/adflib/adf_info.html
// Also credits to Keith Monahan https://www.techtravels.org/tag/mfm/ regarding a bug in the MFM sector start data
//
// V2.8

#pragma once
#include <functional>

#include "RotationExtractor.h"
#include "GreaseweazleInterface.h"


namespace ArduinoFloppyReader {
	enum class ADFResult {
		adfrComplete,					// Process completed successfully
		adfrAborted,					// Process was aborted
		adfrFileError,					// Error opening the file to write to
		adfrFileIOError,				// Error writing to the ADF file
		adfrCompletedWithErrors,		// Completed but with sector errors
		adfrDiskWriteProtected,         // Disk is write protected!
		adfrDriveError,					// Something wrong with reading the disk
		adfrFirmwareTooOld,				// Firmware is too old
		adfrMediaSizeMismatch,			// HD/DD mismatch
		adfrExtendedADFNotSupported,	// Extended ADFs not currently supported
		adfrBadSCPFile,                 // Incompatiable IPF library
		adfrIPFLibraryNotAvailable      // The IPF library was not found
	};

	enum class WriteResponse {
		wrContinue,				// Continue working as normal
		wrAbort,				// Abort thr process and stop
		wrSkipBadChecksums,		// Request that the process ignores bad checksums (not recommended unless the retryCounter gets beyond RETRYS_PER_PHASE*16)
		wrRetry                 // Retry!
	};

	enum class CallbackOperation {
		coStarting,
		coReading,
		coWriting,
		coVerifying,
		coRetryReading,
		coRetryWriting,
		coReVerifying,
		coReadingFile
	};

	// Main writer class
	class ADFWriter {
	private:
		// The Arduino device
		GreaseWeazleInterface m_device;
		bool m_isInHDMode = false;

	public:  
		ADFWriter();
		~ADFWriter();

		inline GreaseWeazleInterface& getDevice() { return m_device; };

		// Open the device we want to use.  Returns TRUE if it worked
		bool openDevice(const std::wstring& portName, bool useDriveB);

		// Close the device when we've finished
		void closeDevice();

		// Reads the disk and write the data to the ADF file supplied.  The callback is for progress, and you can returns FALSE to abort the process
		ADFResult DiskToADF(const std::wstring& outputFile, const bool inHDMode, const unsigned int numTracks, std::function < WriteResponse(const int currentTrack, const DiskSurface currentSide, const int retryCounter, const int sectorsFound, const int badSectorsFound, const int maxSectors, const CallbackOperation operation)> callback);
		
		ADFResult DiskToSCP(const std::wstring& outputFile, const unsigned int numTracks, const unsigned char revolutions, float rpm, std::function < WriteResponse(const int currentTrack, const DiskSurface currentSide, const unsigned int trackTimeMS, const CallbackOperation operation)> callback);

		bool diskClean(std::function<void(int progress)> onProgress);

		bool fetchDiskLabel(bool inHDMode, std::string& label);
	};
};