/* Greaseweazle C++ Interface for reading and writing Amiga Disks
*
* Copyright (C) 2021-2023 Robert Smith (@RobSmithDev)
* https://amiga.robsmithdev.co.uk
*
* This file is multi-licensed under the terms of the Mozilla Public
* License Version 2.0 as published by Mozilla Corporation and the
* GNU General Public License, version 2 or later, as published by the
* Free Software Foundation.
*
* MPL2: https://www.mozilla.org/en-US/MPL/2.0/
* GPL2: https://www.gnu.org/licenses/old-licenses/gpl-2.0.en.html
*
* Based on the excellent code by Keir Fraser <keir.xen@gmail.com>
* https://github.com/keirf/Greaseweazle/
* 
* This file, along with currently active and supported interfaces
* are maintained from by GitHub repo at
* https://github.com/RobSmithDev/FloppyDriveBridge
*/

#include <sstream>
#include <vector>
#include <queue>
#include <cstring>
#include "GreaseWeazleInterface.h"
#include "RotationExtractor.h"
#include "pll.h"

#define ONE_NANOSECOND 1000000000UL
#define BITCELL_SIZE_IN_NS 2000L

#define PIN_DISKCHG_IBMPC	34
#define PIN_DISKCHG_SHUGART 2     // Greaseweazle cannot read from this pin
#define PIN_WRPROT			28


enum class GetInfo { Firmware = 0, BandwidthStats = 1 };
// ## Cmd.{Get,Set}Params indexes
enum class Params { Delays = 0 };
// ## Flux read stream opcodes, preceded by 0xFF byte
enum class FluxOp { Index = 1, Space = 2, Astable = 3};

#pragma pack(1) 

/* CMD_READ_FLUX */
struct GWReadFlux {
	/* Maximum ticks to read for (or 0, for no limit). */
	uint32_t ticks;
	/* Maximum index pulses to read (or 0, for no limit). */
	uint16_t max_index;
	/* Linger time, in ticks, to continue reading after @max_index pulses. */
	uint32_t max_index_linger; /* default: 500 microseconds */
};

/* CMD_WRITE_FLUX */
struct GWWriteFlux {
	/* If non-zero, start the write at the index pulse. */
	unsigned char cue_at_index;
	/* If non-zero, terminate the write at the next index pulse. */
	unsigned char terminate_at_index;
};

#pragma pack()

struct Sequence {
	unsigned char sequence;
	uint16_t speed;     // speed compared to what it should be (for this sequence)
	bool atIndex;			// At index
};

// We're working in nanoseconds
struct PLLData {
	uint32_t freq = 0;			  // sample frequency in Hz
	uint32_t ticks = 0;
	bool indexHit = false;
};


// Change the drive delays as required
bool GreaseWeazleInterface::updateDriveDelays() {
	unsigned char buffer[11];
	buffer[0] = (unsigned char)Params::Delays;
#ifdef _WIN32	
	memcpy_s(buffer + 1, 10, &m_gwDriveDelays, sizeof(m_gwDriveDelays));
#else
	memcpy(buffer + 1, &m_gwDriveDelays, sizeof(m_gwDriveDelays));
#endif	

	Ack response = Ack::Okay;
	if (!sendCommand(Cmd::SetParams, buffer, 11, response)) return false;

	return true;
}

// Constructor for this class
GreaseWeazleInterface::GreaseWeazleInterface() : m_currentBusType(BusType::IBMPC), m_currentDriveIndex(0) {
	memset(&m_gwVersionInformation, 0, sizeof(m_gwVersionInformation));
	memset(&m_gwDriveDelays, 0, sizeof(m_gwDriveDelays));
}

// Free me
GreaseWeazleInterface::~GreaseWeazleInterface() {
	closePort();
}

// Attempts to open the reader running on the COM port provided (or blank for auto-detect)
GWResponse GreaseWeazleInterface::openPort(const std::wstring& comPort, bool driveB) {
	closePort();

	m_motorIsEnabled = false;

	m_inHDMode = false;

	switch (m_comPort.openPort(comPort)) {
	case SerialIO::Response::rInUse:return GWResponse::drPortInUse;
	case SerialIO::Response::rNotFound:return GWResponse::drPortNotFound;
	case SerialIO::Response::rOK: break;
	default: return GWResponse::drPortError;
	}

	// Configure the port
	SerialIO::Configuration config;
	config.baudRate = 9600;
	config.ctsFlowControl = false;

	if (m_comPort.configurePort(config) != SerialIO::Response::rOK) return GWResponse::drPortError;

	applyCommTimeouts(false);

	m_comPort.purgeBuffers();

	// Lets ask GW it's firmware version
	Ack response = Ack::Okay;
	if (!sendCommand(Cmd::GetInfo, (unsigned char)GetInfo::Firmware, response)) {
		m_comPort.purgeBuffers();

		if (!sendCommand(Cmd::GetInfo, (unsigned char)GetInfo::Firmware, response)) {
			closePort();
			return GWResponse::drErrorMalformedVersion;
		}
	};
	unsigned long read = m_comPort.read(&m_gwVersionInformation, sizeof(m_gwVersionInformation));
	if (read != 32) {
		closePort();
		return GWResponse::drErrorMalformedVersion;
	}

	// Reset it
	if (!sendCommand(Cmd::Reset, nullptr, 0, response)) {
		closePort();
		return GWResponse::drErrorMalformedVersion;
	};

	// Now get the drive delays
	if (!sendCommand(Cmd::GetParams, (unsigned char)Params::Delays, response, sizeof(m_gwDriveDelays))) {
		closePort();
		return GWResponse::drErrorMalformedVersion;
	};

	read = m_comPort.read(&m_gwDriveDelays, sizeof(m_gwDriveDelays));
	if (read != sizeof(m_gwDriveDelays)) {
		closePort();
		return GWResponse::drErrorMalformedVersion;
	};

	// Select the drive we want to communicate with
	m_currentBusType = BusType::IBMPC;
	m_currentDriveIndex = driveB?1:0;
	
	if (!sendCommand(Cmd::SetBusType, (unsigned char)m_currentBusType, response)) {
		closePort();
		return GWResponse::drError;
	}

	

	// Ok, success
	return GWResponse::drOK;
}

// Trigger drive select
bool GreaseWeazleInterface::selectDrive(bool select) {
	Ack response = Ack::Okay;

	if (select == m_selectStatus) return true;

	if (select) {
		if (sendCommand(Cmd::Select, m_currentDriveIndex, response)) {
			m_selectStatus = select;
			return true;
		}
	}
	else {
		if (sendCommand(Cmd::Deselect, nullptr, 0, response)) {
			m_selectStatus = select;
			return true;
		}
	}
	return false;
}

// Apply and change the timeouts on the com port
void GreaseWeazleInterface::applyCommTimeouts(bool shortTimeouts) {
	m_comPort.setWriteTimeouts(2000, 200);

	if (shortTimeouts) {
		m_comPort.setReadTimeouts(15, 5);
	}
	else {
		m_comPort.setReadTimeouts(2000, 200);
	}		
}

// Closes the port down
void GreaseWeazleInterface::closePort() {
	enableMotor(false);
	selectDrive(false);
	
	m_comPort.closePort();
}

// Turns on and off the reading interface
GWResponse GreaseWeazleInterface::enableMotor(const bool enable, const bool dontWait) {
	if (!m_comPort.isPortOpen()) return GWResponse::drPortError;
	Ack response = Ack::Okay;

	std::vector<unsigned char> params;
	
	unsigned int delay = dontWait ? 10 : 750;
	if (delay != m_gwDriveDelays.motor_delay) {
		m_gwDriveDelays.motor_delay = delay;
		updateDriveDelays();
	}

	unsigned char buffer[2] = { m_currentDriveIndex, ((unsigned char)(enable ? 1 : 0))};
	if (!sendCommand(Cmd::Motor, buffer, 2, response)) {
		return GWResponse::drError;
	}

	if (response == Ack::Okay) {
		m_motorIsEnabled = enable;
		if (m_motorIsEnabled) selectDrive(true);
	}

	return (response == Ack::Okay) ? GWResponse::drOK : GWResponse::drError;
}


// Select the track, this makes the motor seek to this position
GWResponse GreaseWeazleInterface::selectTrack(const unsigned char trackIndex, const TrackSearchSpeed searchSpeed) {
	if (trackIndex > 81) {
		return GWResponse::drTrackRangeError; // no chance, it can't be done.
	}

	uint16_t newSpeed = m_gwDriveDelays.step_delay;
	switch (searchSpeed) {
		case TrackSearchSpeed::tssSlow:		newSpeed = 5000;  break;
		case TrackSearchSpeed::tssNormal:	newSpeed = 3000;  break;
		case TrackSearchSpeed::tssFast:		newSpeed = 3000;  break;
		case TrackSearchSpeed::tssVeryFast: newSpeed = 3000;  break;
	}
	if (newSpeed != m_gwDriveDelays.step_delay) {
		m_gwDriveDelays.step_delay = newSpeed;
		updateDriveDelays();
	}	

	selectDrive(true);

	Ack response = Ack::Okay;
	sendCommand(Cmd::Seek, trackIndex, response);
	if (!m_motorIsEnabled) selectDrive(false);

	switch (response) {
	case Ack::NoTrk0: return GWResponse::drRewindFailure;
	case Ack::Okay: return GWResponse::drOK;
	default: return GWResponse::drSelectTrackError;
	}
}

// Search for track 0
GWResponse GreaseWeazleInterface::findTrack0() {
	return selectTrack(0, TrackSearchSpeed::tssFast);
}

// Choose which surface of the disk to read from
GWResponse GreaseWeazleInterface::selectSurface(const DiskSurface side) {
	Ack response = Ack::Okay;
	sendCommand(Cmd::Head, (side == DiskSurface::dsUpper ? 1 : 0), response);

	return (response == Ack::Okay) ? GWResponse::drOK : GWResponse::drError;
}

// send a command out to the GW and receive its response.  Returns FALSE on error
bool GreaseWeazleInterface::sendCommand(Cmd command, void* params, unsigned int paramsLength, Ack& response, unsigned char extraResponseSize) {
	if (!m_comPort.isPortOpen()) return false;
	std::vector<unsigned char> data;
	data.resize(paramsLength + 2);
	data[0] = (unsigned char)command;
	data[1] = 2 + paramsLength + (extraResponseSize>0 ? 1 : 0);
	if ((params) && (paramsLength)) {
#ifdef _WIN32		
		memcpy_s(((unsigned char*)data.data()) + 2, paramsLength, params, paramsLength);
#else
		memcpy(((unsigned char*)data.data()) + 2, params, paramsLength);
#endif		
	}
	if (extraResponseSize>0) data.push_back(extraResponseSize);

	// Write request
	unsigned long written = (unsigned long)m_comPort.write(data.data(), (unsigned int)data.size());
	if (written != data.size()) {
		response = Ack::BadCommand;
		return false;
	}
	
	// Read response
	unsigned char responseMsg[2];
	unsigned long read = m_comPort.read(&responseMsg, 2);
	if (read != 2) {
		if (read == 0) {
			// Try again
			read = m_comPort.read(&responseMsg, 2);
		}
		if (read != 2) {
			response = Ack::BadCommand;
			return false;
		}
	}

	response = (Ack)responseMsg[1];

	// Was it the response to the command we issued?
	if (responseMsg[0] != (unsigned char)command) {
		response = Ack::BadCommand;
		return false;
	}

	return true;
}
bool GreaseWeazleInterface::sendCommand(Cmd command, const std::vector<unsigned char>& params, Ack& response, unsigned char extraResponseSize) {
	return sendCommand(command, (void*)params.data(), (unsigned int)params.size(), response, extraResponseSize);
}
bool GreaseWeazleInterface::sendCommand(Cmd command, unsigned char param, Ack& response, unsigned char extraResponseSize) {
	return sendCommand(command, (void*)(&param), 1, response, extraResponseSize);
}

// Taken from optimised.c and modified
static int read_28bit(std::queue<unsigned char>& queue) {
	int x;
	x = queue.front() >> 1;  queue.pop();
	x |= (queue.front() & 0xfe) << 6; queue.pop();
	x |= (queue.front() & 0xfe) << 13; queue.pop();
	x |= (queue.front() & 0xfe) << 20; queue.pop();

	return x;
}

// Convert from nSec to TICKS
static unsigned int nSecToTicks(unsigned int nSec, unsigned int sampleFrequency) {
	return ((unsigned long long)nSec * (unsigned long long)sampleFrequency) / (unsigned long long)ONE_NANOSECOND;
}

// Convert from TICKS to nSec
static unsigned int ticksToNSec(unsigned int ticks, unsigned int sampleFrequency) {
	return ((unsigned long long)ticks * ONE_NANOSECOND) / (unsigned long long)sampleFrequency;
}

// Look at the stream and count up the types of data
void countSampleTypes(PLLData& pllData, std::queue<unsigned char>& queue, unsigned int& hdBits, unsigned int& ddBits) {
	while (!queue.empty()) {

		unsigned char i = queue.front();
		if (i == 255) {
			if (queue.size() < 6) return;
			queue.pop();

			// Get opcode
			switch ((FluxOp)queue.front()) {
			case FluxOp::Index:
				queue.pop();
				read_28bit(queue);
				break;

			case FluxOp::Space:
				queue.pop();
				pllData.ticks += read_28bit(queue);
				break;

			default:
				queue.pop();
				break;
			}
		}
		else {
			if (i < 250) {
				pllData.ticks += (int)i;
				queue.pop();
			}
			else {
				if (queue.size() < 2) return;
				queue.pop();
				pllData.ticks += (250 + (((unsigned int)i) - 250) * 255) + (queue.front() - 1);
				queue.pop();
			}

			// Work out the actual time this tick took in nanoSeconds.
			unsigned int tickInNS = ticksToNSec(pllData.ticks, pllData.freq);
			if (tickInNS > BITCELL_SIZE_IN_NS) {
				if (tickInNS < 3000) hdBits++; 
				if ((tickInNS > 4500) && (tickInNS < 8000)) ddBits++;
				pllData.ticks = 0;
			}
		}
	}
}

// Test the inserted disk and see if its HD or not
GWResponse GreaseWeazleInterface::checkDiskCapacity(bool& isHD) {
	GWReadFlux header{};
	header.ticks = 0;
	header.max_index = 1;
	header.max_index_linger = nSecToTicks(50, m_gwVersionInformation.sample_freq);

	bool alreadySpun = m_motorIsEnabled;
	if (!alreadySpun)
		if (enableMotor(true, false) != GWResponse::drOK) return GWResponse::drError;

	selectDrive(true);

	// Write request
	Ack response = Ack::Okay;
	if (!sendCommand(Cmd::ReadFlux, (void*)&header, sizeof(header), response)) {
		selectDrive(false);
		return GWResponse::drError;
	}

	// Buffer to read into
	unsigned char tempReadBuffer[64] = { 0 };
	bool zeroDetected = false;
	unsigned int failCount = 0;

	// Read the flux data
	PLLData pllData;
	pllData.freq = m_gwVersionInformation.sample_freq;
	unsigned int hdBits = 0;
	unsigned int ddBits = 0;

	std::queue<unsigned char> queue;

	do {
		// More efficient to read several bytes in one go		
		unsigned long bytesAvailable = m_comPort.getBytesWaiting();
		if (bytesAvailable < 1) bytesAvailable = 1;
		if (bytesAvailable > sizeof(tempReadBuffer)) bytesAvailable = sizeof(tempReadBuffer);
		unsigned long bytesRead = m_comPort.read(tempReadBuffer, m_shouldAbortReading ? 1 : bytesAvailable);
		if (bytesRead < 1) {
			failCount++;
			if (failCount > 10) break;
		}
		else failCount = 0;

		if (bytesRead) {

			for (unsigned int a = 0; a < bytesRead; a++) {
				queue.push(tempReadBuffer[a]);
				zeroDetected |= tempReadBuffer[a] == 0;
			}

			// Look at the stream and count up the types of data
			countSampleTypes(pllData, queue, hdBits, ddBits);
		}		
	} while (!zeroDetected);

	// Check for errors
	response = Ack::Okay;
	sendCommand(Cmd::GetFluxStatus, nullptr, 0, response);

	selectDrive(false);

	if (response == Ack::NoIndex) return GWResponse::drNoDiskInDrive;

	if (!alreadySpun) enableMotor(false, false);

	isHD = (hdBits > ddBits);
	return GWResponse::drOK;
}

// Read a bit from the MFM data provided
inline int readBit(const unsigned char* buffer, const unsigned int maxLength, int& pos, int& bit) {
	if (pos >= (int)maxLength) {
		bit--;
		if (bit < 0) {
			bit = 7;
			pos++;
		}
		return (bit & 1) ? 0 : 1;
	}

	int ret = (buffer[pos] >> bit) & 1;
	bit--;
	if (bit < 0) {
		bit = 7;
		pos++;
	}

	return ret;
}


// Process queue and work out whats going on
inline void unpackStreamQueue(std::queue<unsigned char>& queue, PLLData& pllData, PLL::BridgePLL& pll, bool isHDMode) {
	// Run until theres not enough data left to process
	while (!queue.empty()) {

		unsigned char i = queue.front();
		if (i == 255) {
			if (queue.size() < 6) return;
			queue.pop();

			// Get opcode
			switch ((FluxOp)queue.front()) {
			case FluxOp::Index:
				queue.pop();
				read_28bit(queue);
				pllData.indexHit = true;
				break;

			case FluxOp::Space:
				queue.pop();
				pllData.ticks += read_28bit(queue);
				break;

			default:
				queue.pop();
				break;
			}
		}
		else {
			if (i < 250) {
				pllData.ticks += (int32_t)i;
				queue.pop();
			}
			else {
				if (queue.size() < 2) return;
				queue.pop();				
				pllData.ticks += (250 + (((int32_t)i) - 250) * 255) + ((int32_t)queue.front() - 1);
				queue.pop();
			}

			// Work out the actual time this tick took in nanoSeconds.
			uint32_t tickInNS = ticksToNSec(pllData.ticks, pllData.freq);

			// This is how the Amiga expects it
			if (isHDMode) tickInNS *= 2;

			pll.submitFlux(tickInNS, pllData.indexHit);
			pllData.indexHit = false;
			pllData.ticks = 0;
		}
	}
}


// Process queue and work out whats going on
inline void unpackStreamQueue2(std::queue<unsigned char>& queue, PLLData& pllData, unsigned char* output, int& dataPosition, int maxLength, bool isHDMode) {
	// Run until theres not enough data left to process
	while (!queue.empty()) {

		unsigned char i = queue.front();
		if (i == 255) {
			if (queue.size() < 6) return;
			queue.pop();

			// Get opcode
			switch ((FluxOp)queue.front()) {
			case FluxOp::Index:
				queue.pop();
				read_28bit(queue);
				pllData.indexHit = true;
				break;

			case FluxOp::Space:
				queue.pop();
				pllData.ticks += read_28bit(queue);
				break;

			default:
				queue.pop();
				break;
			}
		}
		else {
			if (i < 250) {
				pllData.ticks += (int32_t)i;
				queue.pop();
			}
			else {
				if (queue.size() < 2) return;
				queue.pop();
				pllData.ticks += (250 + (((int32_t)i) - 250) * 255) + ((int32_t)queue.front() - 1);
				queue.pop();
			}

			// Work out the actual time this tick took in nanoSeconds.
			uint32_t tickInNS = ticksToNSec(pllData.ticks, pllData.freq);

			// This is how the Amiga expects it
			if (isHDMode) tickInNS *= 2;

			int bits = (tickInNS+1000) / 2000;


			if (tickInNS > BITCELL_SIZE_IN_NS) {

				int t = tickInNS - BITCELL_SIZE_IN_NS;

				// Clock out the sequence
				while (t > (BITCELL_SIZE_IN_NS / 2)) {
					if (dataPosition < maxLength) {
						output[dataPosition / 8] <<= 1;
						dataPosition++;
					}
					t -= BITCELL_SIZE_IN_NS;
				}
				if (dataPosition < maxLength) {
					output[dataPosition / 8] <<= 1;
					output[dataPosition / 8] |= 1;
					dataPosition++;
				}

				// Save it
				pllData.ticks = 0;
			}
			pllData.indexHit = false;
		}
	}
}


// Process queue and work out whats going on
inline void unpackStreamQueue3(std::queue<unsigned char>& queue, PLLData& pllData, std::vector<DWORD>& fluxTransitions, std::function<void(const std::vector<DWORD>& fluxTransitions, DWORD extraTimeToAddOn)> onRotation, int& numRevolutions) {
	// Run until theres not enough data left to process
	int tmp;
	while (!queue.empty()) {

		unsigned char i = queue.front();
		if (i == 255) {
			if (queue.size() < 6) return;
			queue.pop();

			// Get opcode
			switch ((FluxOp)queue.front()) {
			case FluxOp::Index:
				queue.pop();
				tmp = read_28bit(queue);
				pllData.indexHit = true;
				if (fluxTransitions.size()) {
					if (numRevolutions) {
						numRevolutions--;
						onRotation(fluxTransitions, pllData.ticks + tmp);
					}
				}
				pllData.ticks = 0;
				fluxTransitions.clear();
				break;

			case FluxOp::Space:
				queue.pop();
				pllData.ticks += read_28bit(queue);
				break;

			default:
				queue.pop();
				break;
			}
		}
		else {
			if (i < 250) {
				pllData.ticks += (int32_t)i;
				queue.pop();
			}
			else {
				if (queue.size() < 2) return;
				queue.pop();
				pllData.ticks += (250 + (((int32_t)i) - 250) * 255) + ((int32_t)queue.front() - 1);
				queue.pop();
			}

			// Work out the actual time this tick took in nanoSeconds.
			uint32_t tickInNS = ticksToNSec(pllData.ticks, pllData.freq);
			if (pllData.indexHit) fluxTransitions.push_back(tickInNS);
			pllData.ticks = 0;
		}
	}
}


// Process queue and work out whats going on
inline void unpackStreamQueue4(std::queue<unsigned char>& queue, uint32_t sampleFreq, uint64_t& ticksSoFar, bool& indexFound, std::function<void(DWORD revolutionTime)> onRotation, int &numRevolutions) {
	// Run until theres not enough data left to process
	while (!queue.empty()) {

		unsigned char i = queue.front();
		if (i == 255) {
			if (queue.size() < 6) return;
			queue.pop();

			// Get opcode
			switch ((FluxOp)queue.front()) {
			case FluxOp::Index:
				queue.pop();
				read_28bit(queue);
				if (indexFound && ticksSoFar) {
					if (numRevolutions) {
						numRevolutions--;
						DWORD time = (DWORD)ticksToNSec((unsigned int)ticksSoFar, sampleFreq);
						onRotation(time);
					}
				}
				ticksSoFar = 0;
				indexFound = true;

				break;

			case FluxOp::Space:
				queue.pop();
				ticksSoFar += read_28bit(queue);
				break;

			default:
				queue.pop();
				break;
			}
		}
		else {
			if (i < 250) {
				ticksSoFar += (uint64_t)i;
				queue.pop();
			}
			else {
				if (queue.size() < 2) return;
				queue.pop();
				ticksSoFar += (250 + (((int32_t)i) - 250) * 255) + ((int32_t)queue.front() - 1);
				queue.pop();
			}
		}
	}
}


// Reads a complete rotation of the disk, and returns it using the callback function which can return FALSE to stop
// An instance of RotationExtractor is required.  This is purely to save on re-allocations.  It is internally reset each time
GWResponse GreaseWeazleInterface::readRotation(PLL::BridgePLL& pll, const unsigned int maxOutputSize, RotationExtractor::MFMSample* firstOutputBuffer, RotationExtractor::IndexSequenceMarker& startBitPatterns,
	std::function<bool(RotationExtractor::MFMSample** mfmData, const unsigned int dataLengthInBits)> onRotation) {	
	GWReadFlux header;

	const uint32_t extraTime = OVERLAP_SEQUENCE_MATCHES * (OVERLAP_EXTRA_BUFFER) * 8000;
	if ((pll.rotationExtractor()->isInIndexMode()) || (!pll.rotationExtractor()->hasLearntRotationSpeed())) {
		header.ticks = 0;
		header.max_index = 1;
		header.max_index_linger = nSecToTicks((212 * 1000 * 1000) + extraTime, m_gwVersionInformation.sample_freq);
	}
	else {
		header.ticks = nSecToTicks((212 * 1000 * 1000) + extraTime, m_gwVersionInformation.sample_freq);
		header.max_index = 0;
		header.max_index_linger = 0;
	}
	
	m_shouldAbortReading = false;

	// Sample storage
	std::queue<unsigned char> queue;

	// Reset ready for extraction
	pll.prepareExtractor(m_inHDMode, startBitPatterns);

	selectDrive(true);

	// Write request
	Ack response = Ack::Okay;
	if (!sendCommand(Cmd::ReadFlux, (void*)&header, sizeof(header), response)) {
		if (!m_motorIsEnabled) selectDrive(false);
		return GWResponse::drReadResponseFailed;
	}
	int32_t failCount = 0;

	// Buffer to read into
	unsigned char tempReadBuffer[64] = { 0 };
	bool zeroDetected = false;

	applyCommTimeouts(true);

	PLLData pllData;
	pllData.freq = m_gwVersionInformation.sample_freq;

	do {
		// More efficient to read several bytes in one go		
		uint32_t bytesAvailable = m_comPort.getBytesWaiting();
		if (bytesAvailable < 1) bytesAvailable = 1;
		if (bytesAvailable > sizeof(tempReadBuffer)) bytesAvailable = sizeof(tempReadBuffer);
		uint32_t bytesRead = m_comPort.read(tempReadBuffer, m_shouldAbortReading ? 1 : bytesAvailable);
		if (bytesRead<1) {
			failCount++;
			if (failCount > 10) break;
		}
		else failCount = 0;

		// If theres this many we can process as this is the maximum we need to process
		if ((!m_shouldAbortReading) && (bytesRead)) {

			for (uint32_t a = 0; a < bytesRead; a++) {
				queue.push(tempReadBuffer[a]);
				zeroDetected |= tempReadBuffer[a] == 0;
			}

			unpackStreamQueue(queue, pllData, pll, m_inHDMode);

			// Is it ready to extract?
			if (pll.canExtract()) {
				uint32_t bits = 0;
				// Go!
				if (pll.extractRotation(firstOutputBuffer, bits, maxOutputSize)) {
					if (!onRotation(&firstOutputBuffer, bits)) {
						// And if the callback says so we stop. - unsupported at present
						abortReadStreaming();
					}
					// Always save this back
					pll.getIndexSequence(startBitPatterns);
				}
			}
		}
		else {
			for (uint32_t a = 0; a < bytesRead; a++) zeroDetected |= tempReadBuffer[a] == 0;
		}
	} while (!zeroDetected);

	applyCommTimeouts(false);

	// Check for errors
	response = Ack::Okay;
	sendCommand(Cmd::GetFluxStatus, nullptr, 0, response);

	if (!m_motorIsEnabled) selectDrive(false);
	switch (response) {
	case Ack::FluxOverflow: return GWResponse::drSerialOverrun;
	case Ack::NoIndex: return GWResponse::drNoDiskInDrive;
	case Ack::Okay: return GWResponse::drOK;
	default: return  GWResponse::drReadResponseFailed;
	}
}


// Get the disk RPM
GWResponse GreaseWeazleInterface::calculateRPM(float& rpm) {
	GWReadFlux header;
	rpm = 0;
	header.ticks = 0;
	header.max_index = 5;
	header.max_index_linger = 0;

	m_shouldAbortReading = false;

	// Sample storage
	std::queue<unsigned char> queue;
	selectDrive(true);

	// Write request
	Ack response = Ack::Okay;
	if (!sendCommand(Cmd::ReadFlux, (void*)&header, sizeof(header), response)) {
		if (!m_motorIsEnabled) selectDrive(false);
		return GWResponse::drReadResponseFailed;
	}
	int32_t failCount = 0;

	// Buffer to read into
	unsigned char tempReadBuffer[64] = { 0 };
	bool zeroDetected = false;

	applyCommTimeouts(true);
	int revs = 5;
	uint64_t totalTime = 0;	
	int revCount = 0;
	int outPosition = 0;
	uint64_t ticksSoFar = 0;
	bool indexFound = false;

	do {
		// More efficient to read several bytes in one go		
		uint32_t bytesAvailable = m_comPort.getBytesWaiting();
		if (bytesAvailable < 1) bytesAvailable = 1;
		if (bytesAvailable > sizeof(tempReadBuffer)) bytesAvailable = sizeof(tempReadBuffer);
		uint32_t bytesRead = m_comPort.read(tempReadBuffer, m_shouldAbortReading ? 1 : bytesAvailable);
		if (bytesRead < 1) {
			failCount++;
			if (failCount > 10) break;
		}
		else failCount = 0;

		// If theres this many we can process as this is the maximum we need to process
		if ((!m_shouldAbortReading) && (bytesRead)) {

			for (uint32_t a = 0; a < bytesRead; a++) {
				queue.push(tempReadBuffer[a]);
				zeroDetected |= tempReadBuffer[a] == 0;
			}

			unpackStreamQueue4(queue, m_gwVersionInformation.sample_freq, ticksSoFar, indexFound, [this, &totalTime, &revCount](DWORD time) {
				totalTime += time;
				revCount++;
			}, revs);
		}
		else {
			for (uint32_t a = 0; a < bytesRead; a++) zeroDetected |= tempReadBuffer[a] == 0;
		}
	} while (!zeroDetected);

	if (revCount) {
		rpm = (float)(1.0 / ((((double)(totalTime) / (double)revCount))/60000000000.0));
	}

	applyCommTimeouts(false);

	// Check for errors
	response = Ack::Okay;
	sendCommand(Cmd::GetFluxStatus, nullptr, 0, response);

	if (!m_motorIsEnabled) selectDrive(false);
	switch (response) {
	case Ack::FluxOverflow: return GWResponse::drSerialOverrun;
	case Ack::NoIndex: return GWResponse::drNoDiskInDrive;
	case Ack::Okay: return GWResponse::drOK;
	default: return  GWResponse::drReadResponseFailed;
	}
}


// Simple version of the below with no PLL
GWResponse GreaseWeazleInterface::readCurrentTrackFlux(int numRevolutions, std::function<void(const std::vector<DWORD>& fluxTransitions, DWORD extraTimeToAddOn)> onRotation) {
	GWReadFlux header;

	header.ticks = 0;
	header.max_index = numRevolutions+1;
	header.max_index_linger = 0;

	m_shouldAbortReading = false;

	// Sample storage
	std::queue<unsigned char> queue;

	selectDrive(true);

	// Write request
	Ack response = Ack::Okay;
	if (!sendCommand(Cmd::ReadFlux, (void*)&header, sizeof(header), response)) {
		if (!m_motorIsEnabled) selectDrive(false);
		return GWResponse::drReadResponseFailed;
	}
	int32_t failCount = 0;

	// Buffer to read into
	unsigned char tempReadBuffer[64] = { 0 };
	bool zeroDetected = false;

	applyCommTimeouts(true);

	PLLData pllData;
	pllData.freq = m_gwVersionInformation.sample_freq;
	pllData.ticks = 0;
	pllData.indexHit = false;

	std::vector<DWORD> fluxTimings;
	fluxTimings.reserve(42000*2);

	int outPosition = 0;
	do {
		// More efficient to read several bytes in one go		
		uint32_t bytesAvailable = m_comPort.getBytesWaiting();
		if (bytesAvailable < 1) bytesAvailable = 1;
		if (bytesAvailable > sizeof(tempReadBuffer)) bytesAvailable = sizeof(tempReadBuffer);
		uint32_t bytesRead = m_comPort.read(tempReadBuffer, m_shouldAbortReading ? 1 : bytesAvailable);
		if (bytesRead < 1) {
			failCount++;
			if (failCount > 10) break;
		}
		else failCount = 0;

		// If theres this many we can process as this is the maximum we need to process
		if ((!m_shouldAbortReading) && (bytesRead)) {

			for (uint32_t a = 0; a < bytesRead; a++) {
				queue.push(tempReadBuffer[a]);
				zeroDetected |= tempReadBuffer[a] == 0;
			}

			unpackStreamQueue3(queue, pllData, fluxTimings, onRotation, numRevolutions);
		}
		else {
			for (uint32_t a = 0; a < bytesRead; a++) zeroDetected |= tempReadBuffer[a] == 0;
		}
	} while (!zeroDetected);

	applyCommTimeouts(false);

	// Check for errors
	response = Ack::Okay;
	sendCommand(Cmd::GetFluxStatus, nullptr, 0, response);

	if (!m_motorIsEnabled) selectDrive(false);
	switch (response) {
	case Ack::FluxOverflow: return GWResponse::drSerialOverrun;
	case Ack::NoIndex: return GWResponse::drNoDiskInDrive;
	case Ack::Okay: return GWResponse::drOK;
	default: return  GWResponse::drReadResponseFailed;
	}
}


// Read RAW data from the current track and surface selected - this works properly with the HD and DD options
// dataLength must be either RAW_TRACKDATA_LENGTH or RAW_TRACKDATA_LENGTH_HD.  If you mismatch the type then the function will return an error
GWResponse GreaseWeazleInterface::readCurrentTrack(void* trackData, const int dataLength) {
	GWReadFlux header;

	header.ticks = nSecToTicks(225 * 1000 * 1000, m_gwVersionInformation.sample_freq);
	header.max_index = 0;
	header.max_index_linger = 0;

	m_shouldAbortReading = false;

	// Sample storage
	std::queue<unsigned char> queue;

	selectDrive(true);

	// Write request
	Ack response = Ack::Okay;
	if (!sendCommand(Cmd::ReadFlux, (void*)&header, sizeof(header), response)) {
		if (!m_motorIsEnabled) selectDrive(false);
		return GWResponse::drReadResponseFailed;
	}
	int32_t failCount = 0;

	// Buffer to read into
	unsigned char tempReadBuffer[64] = { 0 };
	bool zeroDetected = false;

	applyCommTimeouts(true);

	PLLData pllData;
	pllData.freq = m_gwVersionInformation.sample_freq;
	pllData.ticks = 0;
	pllData.indexHit = false;

	int outPosition = 0;	
	do {
		// More efficient to read several bytes in one go		
		uint32_t bytesAvailable = m_comPort.getBytesWaiting();
		if (bytesAvailable < 1) bytesAvailable = 1;
		if (bytesAvailable > sizeof(tempReadBuffer)) bytesAvailable = sizeof(tempReadBuffer);
		uint32_t bytesRead = m_comPort.read(tempReadBuffer, m_shouldAbortReading ? 1 : bytesAvailable);
		if (bytesRead < 1) {
			failCount++;
			if (failCount > 10) break;
		}
		else failCount = 0;

		// If theres this many we can process as this is the maximum we need to process
		if ((!m_shouldAbortReading) && (bytesRead)) {

			for (uint32_t a = 0; a < bytesRead; a++) {
				queue.push(tempReadBuffer[a]);
				zeroDetected |= tempReadBuffer[a] == 0;
			}

			unpackStreamQueue2(queue, pllData, (unsigned char*)trackData, outPosition, dataLength, m_inHDMode);
		}
		else {
			for (uint32_t a = 0; a < bytesRead; a++) zeroDetected |= tempReadBuffer[a] == 0;
		}
	} while (!zeroDetected);

	applyCommTimeouts(false);

	// Check for errors
	response = Ack::Okay;
	sendCommand(Cmd::GetFluxStatus, nullptr, 0, response);

	if (!m_motorIsEnabled) selectDrive(false);
	switch (response) {
	case Ack::FluxOverflow: return GWResponse::drSerialOverrun;
	case Ack::NoIndex: return GWResponse::drNoDiskInDrive;
	case Ack::Okay: return GWResponse::drOK;
	default: return  GWResponse::drReadResponseFailed;
	}
}



// Attempt to abort reading
void GreaseWeazleInterface::abortReadStreaming() {
	m_shouldAbortReading = true;
}

